# Standard library imports
import sys
from email.headerregistry import Address
from logging import getLogger
from os import environ
from pathlib import Path
from typing import Annotated, Self
from zoneinfo import ZoneInfo

# Third party imports
from pydantic import Field
from pydantic_settings import BaseSettings as _BaseSettings, SettingsConfigDict

# First party imports
from aeth_ext.types.abc import CapturesSubclasses

logger = getLogger(__name__)

environ.setdefault("PYDANTIC_ERRORS_INCLUDE_URL", "false")


CWD = Path(__file__).parent if getattr(sys, "frozen", False) else Path.cwd()

__all__ = ["BaseSettings"]

type AddressLike = str | Address | tuple[str, str | None, str | None, str | None]


class BaseSettings(_BaseSettings, CapturesSubclasses):
  model_config = (
    SettingsConfigDict(
      env_file=CWD / ".env",
      env_file_encoding="utf-8",
      env_ignore_empty=True,
      extra="ignore",
    )
    if __debug__
    else SettingsConfigDict()
  )

  persisted_dir_loc: Annotated[Path, Field(alias="PERSISTED_DIR_LOC")] = (
    CWD / "persisted_data" if __debug__ else Path("/app/persisted_data")
  )

  alerts_smtp_server: Annotated[str, Field(alias="ALERTS_SMTP_SERVER")] = "smtppro.zoho.com"
  alerts_smtp_port: Annotated[int, Field(alias="ALERTS_SMTP_PORT")] = 587
  alerts_email: Annotated[AddressLike, Field(alias="ALERTS_EMAIL")] = "info@sweetfiretobacco.com"
  alerts_email_pwd: Annotated[str, Field(alias="ALERTS_EMAIL_PWD")]
  alerts_recipients: Annotated[frozenset[AddressLike], Field(alias="ALERTS_RECIPIENTS")] = frozenset(
    {"jacob.ogden@sweetfiretobacco.com"}
  )

  log_loc_folder: Annotated[Path, Field(alias="LOG_LOC_FOLDER")] = persisted_dir_loc / "logs"

  tz: Annotated[ZoneInfo, Field(alias="TZ")] = ZoneInfo("US/Eastern")

  def _creds_file_reusable(self, err_msg: str, *expected_path_parts: str) -> Path:
    fp = self.persisted_dir_loc.joinpath(*expected_path_parts)
    if not fp.exists() or not fp.is_file():
      raise FileNotFoundError(f"{err_msg}: {fp}")
    return fp

  # Make this an alias of get_final_model to maintain compatibility with existing code that uses get_settings
  @classmethod
  def get_settings(cls) -> Self:
    return cls.get_final_model()  # pyright: ignore[reportReturnType]


if __name__ == "__main__":
  settings = BaseSettings.get_settings()
  pass



# Standard library imports
from abc import abstractmethod
from enum import Enum, auto
from typing import TYPE_CHECKING, NamedTuple, Protocol, override

if TYPE_CHECKING:
  # Standard library imports
  from collections.abc import Buffer, Callable, Iterator
  from datetime import datetime
  from ftplib import FTP
  from io import BytesIO
  from typing import Any

  # Third party imports
  from paramiko import SFTPClient

  # First party imports
  from aeth_ext.ftp.adapter import AdaptedFTP, AdaptedSFTP


__all__ = ["AdapterProtocol", "FTPProtocol", "ListDirResult", "ProtocolEnum", "SFTPProtocol"]

BufferSize = int
TransferSuccess = bool


class ProtocolEnum(Enum):
  FTP = auto()
  SFTP = auto()


class ListDirResult(NamedTuple):
  filename: str
  modified_time: datetime


class FTPProtocolBase(Protocol):
  KIND: ProtocolEnum

  @abstractmethod
  def get_conn_handler(self) -> Any:
    raise NotImplementedError

  @abstractmethod
  def close_conn_handler(self) -> None:
    raise NotImplementedError


class FTPProtocol(FTPProtocolBase):
  KIND = ProtocolEnum.FTP

  @override
  @abstractmethod
  def get_conn_handler(self) -> FTP:
    raise NotImplementedError


class SFTPProtocol(FTPProtocolBase):
  KIND = ProtocolEnum.SFTP

  @override
  @abstractmethod
  def get_conn_handler(self) -> SFTPClient:
    raise NotImplementedError


class AdapterProtocol(Protocol):
  __slots__ = ()

  def test_connection(self, logit: bool = False) -> bool:
    """Tests the connection to the FTP/SFTP server. Returns True if successful, False otherwise."""
    raise NotImplementedError

  def get_size(self, path: str) -> int | None:
    """Expects an absolute path to a file on the FTP/SFTP server and returns its size in bytes."""
    raise NotImplementedError

  def upload_file(self, remote_path: str, callback: Callable[[BufferSize], bytes], file_size: int, task_msg: str = "") -> None:
    """Expects an absolute path to a file on the FTP/SFTP server and returns a writable file-like object (e.g. socket or SFTPFile) that can be used to send the file's contents."""
    raise NotImplementedError

  def download_file(self, remote_path: str, callback: Callable[[Buffer], Any], task_msg: str = "") -> None:
    """Expects an absolute path to a file on the FTP/SFTP server and returns a readable file-like object (e.g. socket or SFTPFile) that can be used to read the file's contents."""
    raise NotImplementedError

  def transfer_file(
    self,
    source_remote_path: str,
    dest_remote_path: str,
    other: AdaptedFTP | AdaptedSFTP,
    task_msg: str = "",
    callback: Callable[[bytes], None] | None = None,
    mem_stream: BytesIO | None = None,
  ) -> TransferSuccess:
    """Transfers a file from source_remote_path to dest_remote_path on the FTP/SFTP server.
    This is intended to be used for server to server transfers that don't save the file locally."""
    raise NotImplementedError

  def rename(self, old_remote_path: str, new_remote_path: str) -> None:
    """Renames a file on the FTP/SFTP server from old_remote_path to new_remote_path."""
    raise NotImplementedError

  def remove(self, remote_path: str) -> None:
    """Removes a file on the FTP/SFTP server at the given absolute path."""
    raise NotImplementedError

  def listdir(self, path: str) -> Iterator[ListDirResult]:
    """Expects an absolute path to a directory on the FTP/SFTP server and returns an iterator of ListDirResult containing the filename and modification time of each file in the directory.
    The filename is not a full path, just the name of the file. The modification time is a datetime object representing the last modification time of the file on the server.
    Note that the modification time may be None if it cannot be determined, and in that case the tuple will not be yielded.
    """
    raise NotImplementedError

  def makedir(self, remote_path: str) -> None:
    """Creates a directory on the FTP/SFTP server at the given absolute path."""
    raise NotImplementedError



# Local folder imports
from .adapter import AdaptedFTP, AdaptedSFTP, FTPAdapter
from .errors import ServerNotAvailableError
from .types import AdapterProtocol, FTPProtocol, ListDirResult, ProtocolEnum, SFTPProtocol

__all__ = [
  "AdaptedFTP",
  "AdaptedSFTP",
  "AdapterProtocol",
  "FTPAdapter",
  "FTPProtocol",
  "ListDirResult",
  "ProtocolEnum",
  "SFTPProtocol",
  "ServerNotAvailableError",
]



# Standard library imports
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any

# Third party imports
from aiohttp import web
from aiohttp.web import FileResponse, Request
from textual_serve.server import Server

if TYPE_CHECKING:
  # Standard library imports
  from os import PathLike


log = logging.getLogger("textual-serve")

_HTTP_PORT = 80
_HTTPS_PORT = 443


class InLoopServer(Server):
  """A server that runs in the current event loop."""

  def __init__(
    self,
    command: str,
    host: str = "localhost",
    port: int = 8000,
    title: str | None = None,
    public_url: str | None = None,
    statics_path: str | PathLike[Any] = "./static",
    templates_path: str | PathLike[Any] = "./templates",
    favicon_path: str | PathLike[Any] = "./favicon.ico",
  ):
    """Initialize the server.

    Args:
        command: Command to run the Textual self.app.
        host: Host of web application.
        port: Port for server.
        title: Title of the self.app.
        public_url: Public URL for the server.
        statics_path: Path to statics folder.
        templates_path: Path to templates folder.
        favicon_path: Path to favicon file.
    """
    super().__init__(
      command=command,
      host=host,
      port=port,
      title=title,
      public_url=public_url,
      statics_path=statics_path,
      templates_path=templates_path,
    )
    base_path = (Path(__file__) / "../").resolve().absolute()
    self.favicon_path = base_path / favicon_path

    self.runner: web.AppRunner | None = None
    self.site: web.TCPSite | None = None
    self.app: web.Application | None = None

  async def favicon(self, request: Request) -> FileResponse:
    return FileResponse(self.favicon_path)

  async def serve_in_loop(self, debug: bool = False) -> web.AppRunner:
    """Serve the Textual application in an already-running event loop.

    Args:
        debug: Enable debug mode for Textual dev tools.
    Returns:
        The aiohttp AppRunner instance.
    """
    self.debug = debug
    self.initialize_logging()

    if self.debug:
      log.info("Running in debug mode. You may use textual dev tools.")

    self.app = await self._make_app()

    self.app.router.add_get("/favicon.ico", self.favicon)
    self.runner = web.AppRunner(self.app)
    await self.runner.setup()
    self.site = web.TCPSite(self.runner, self.host, self.port)
    await self.site.start()
    log.info("Textual server running on http://%s:%d", self.host, self.port)

    return self.runner



# Standard library imports
from functools import partial
from logging import getLogger
from threading import RLock
from typing import TYPE_CHECKING, cast, override

# Third party imports
from rich import get_console
from rich.live import Live
from rich.progress import (
  BarColumn,
  MofNCompleteColumn,
  Progress as _Progress,
  ProgressColumn,
  Task,
  TaskProgressColumn,
  TextColumn,
  TimeRemainingColumn,
)

if TYPE_CHECKING:
  # Standard library imports
  from types import TracebackType
  from typing import Any, Self

  # Third party imports
  from rich.console import Console
  from rich.progress import GetTimeCallable, TaskID as _TaskID

logger = getLogger(__name__)


__all__ = ["Progress", "TaskID"]


class TaskID(int):
  def __new__(cls, task_id: int, prog_instance: _Progress | type[_Progress], remove: bool = True):
    return super().__new__(cls, task_id)

  @override
  def __init__(self, task_id: int, prog_instance: _Progress | type[_Progress], remove: bool = True):
    self.prog_instance = prog_instance
    self.remove = remove
    self.remove_func = partial(prog_instance.remove_task, self)
    super().__init__()

  def __enter__(self) -> Self:
    return self

  def __exit__(
    self,
    exc_type: type[BaseException] | None,
    exc_value: BaseException | None,
    traceback: TracebackType | None,
  ):
    remove = self.remove
    if remove:
      self.remove_func()

  def __copy__(self) -> Self:
    return self

  def __deepcopy__(self, memo: dict[int, Any]) -> Self:
    return self


class Progress(_Progress):
  @override
  def __init__(
    self,
    *columns: str | ProgressColumn,
    console: Console | None,
    auto_refresh: bool = True,
    refresh_per_second: float = 4,
    speed_estimate_period: float = 30.0,
    transient: bool = False,
    redirect_stdout: bool = True,
    redirect_stderr: bool = True,
    get_time: GetTimeCallable | None = None,
    disable: bool = False,
    expand: bool = False,
    live: Live | None = None,
  ) -> None:
    assert refresh_per_second > 0, "refresh_per_second must be > 0"
    if console is None:
      console = get_console()
    self._lock = RLock()
    self.columns = (
      BarColumn(),
      TaskProgressColumn(),
      MofNCompleteColumn(),
      TimeRemainingColumn(),
      TextColumn("[progress.description]{task.description}"),
      *columns,
    )
    self.speed_estimate_period = speed_estimate_period

    self.disable = disable
    self.expand = expand
    self._tasks: dict[TaskID, Task] = {}  # type: ignore
    self._task_index: TaskID = TaskID(0, self)
    self.live = live or Live(
      console=console,
      auto_refresh=auto_refresh,
      refresh_per_second=refresh_per_second,
      transient=transient,
      redirect_stdout=redirect_stdout,
      redirect_stderr=redirect_stderr,
      get_renderable=self.get_renderable,
    )
    self.get_time = get_time or self.console.get_time
    self.print = self.console.print
    self.log = self.console.log

  def add_task(  # type: ignore
    self,
    description: str,
    start: bool = True,
    total: float | None = 100.0,
    completed: int = 0,
    visible: bool = True,
    remove_when_finished: bool = True,
    **fields: Any,
  ) -> TaskID:
    """Add a new 'task' to the Progress display.

    Args:
        description (str): A description of the task.
        start (bool, optional): Start the task immediately (to calculate elapsed time). If set to False,
            you will need to call `start` manually. Defaults to True.
        total (float, optional): Number of total steps in the progress if known.
            Set to None to render a pulsing animation. Defaults to 100.
        completed (int, optional): Number of steps completed so far. Defaults to 0.
        visible (bool, optional): Enable display of the task. Defaults to True.
        **fields (str): Additional data fields required for rendering.

    Returns:
        TaskID: An ID you can use when calling `update`.
    """
    with self._lock:
      task = Task(
        self._task_index,  # type: ignore
        description,
        total,
        completed,
        visible=visible,
        fields=fields,
        _get_time=self.get_time,
        _lock=self._lock,
      )
      self._tasks[self._task_index] = task
      if start:
        self.start_task(self._task_index)  # type: ignore

      new_task_index = self._task_index
      new_task_index.remove = remove_when_finished
      self._task_index = TaskID(int(self._task_index) + 1, self)  # type: ignore
    # self.refresh()
    return new_task_index

  def update(  # type: ignore
    self,
    task_id: TaskID,
    *,
    total: float | None = None,
    completed: float | None = None,
    advance: float | None = None,
    description: str | None = None,
    visible: bool | None = None,
    refresh: bool = False,
    **fields: Any,
  ) -> None:
    return super().update(
      cast("_TaskID", task_id),
      total=total,
      completed=completed,
      advance=advance,
      description=description,
      visible=visible,
      refresh=refresh,
      **fields,
    )

  def remove_task(self, task_id: TaskID) -> None:  # type: ignore
    return super().remove_task(cast("_TaskID", task_id))
